package de.hft.stuttgart.citydoctor2.healing;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.check.HealingID;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.check.error.MultipleConnectedComponentsError;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.healing.math.EdgeGraph;
import de.hft.stuttgart.citydoctor2.math.Plane;

public class GeometrySimplifier implements HealingMethod {

	private static final Logger logger = LogManager.getLogger(GeometrySimplifier.class);

	private static final double EPSILON = 0.1;
	
	@Override
	public HealingID getID() {
		return HealingID.S_GEOMETRIC_SIMPLIFIER;
	}

	@Override
	public boolean visit(MultipleConnectedComponentsError err, ModificationListener l) {
		logger.debug("Executing Repair for MultipleConnectedComponentsError");
		// first merge similar polygons
		for (List<Polygon> component : err.getComponents()) {
			for (int i = 0; i < component.size() - 1; i++) {
				Polygon p0 = component.get(i);
				Plane plane = new Plane(p0.calculateNormalNormalized(), p0.getExteriorRing().getVertices().get(0));
				for (int j = i + 1; j < component.size(); j++) {
					Polygon p1 = component.get(j);
					if (isPolygonInPlane(plane, p1) && isPolygonConnectedViaEdge(p0, p1, err.getGeometry())) {
						ConcretePolygon p = mergePolygons(p0, p1);
						p1.remove();
						err.getGeometry().replacePolygon(p0, p);
						return true;
					}
				}
			}
		}
		return false;
	}
	
	private boolean isPolygonConnectedViaEdge(Polygon p1, Polygon p2, Geometry g) {
		List<Edge> edges = g.getEdgesAdjacentTo(p1);
		for (Edge e : edges) {
			if (e.getAdjacentPolygons().contains(p2)) {
				return true;
			}
		}
		return false;
	}
	
	private ConcretePolygon mergePolygons(Polygon p1, Polygon p2) {
		Geometry geom = p1.getParent();
		List<Edge> edgesP1 = geom.getEdgesAdjacentTo(p1);
		List<Edge> edgesP2 = geom.getEdgesAdjacentTo(p2);
		Set<Edge> setOfEdgesFromP1 = new HashSet<>(edgesP1);
		Set<Edge> duplicateEdges = new HashSet<>();
		for (Edge e : edgesP2) {
			if (setOfEdgesFromP1.contains(e)) {
				duplicateEdges.add(e);
			}
		}
		if (duplicateEdges.isEmpty()) {
			throw new IllegalStateException(
					"Trying to merge two polygon which are not connected " + p1.getGmlId() + ", " + p2.getGmlId());
		}
		List<Edge> newEdges = new ArrayList<>();
		filterDuplicateEdges(edgesP1, duplicateEdges, newEdges);
		filterDuplicateEdges(edgesP2, duplicateEdges, newEdges);
		if (newEdges.isEmpty()) {
			return copyPolygon(p1);
		}
		List<List<Edge>> rings = findCyclesInEdges(newEdges);
		// find the exterior ring in the new rings
		ConcretePolygon p = new ConcretePolygon();
		List<LinearRing> newRings = new ArrayList<>();
		for (List<Edge> edges : rings) {
			LinearRing ring = createRingFromEdges(edges);
			newRings.add(ring);
		}
		if (rings.size() == 1) {
			// only one ring -> the ring is an exterior
			LinearRing ext = newRings.get(0);
			ext.setType(LinearRingType.EXTERIOR);

			p.setParent(geom);
			p.setExteriorRing(ext);
			return p;
		} else {
			for (LinearRing lr : newRings) {
				if (isRingExterior(newRings, lr)) {
					return createPolygon(newRings, lr, p);
				}
			}
			throw new IllegalStateException(
					"Merging polygons produced intersecting edges," + " no exterior ring could be found");
		}

	}

	private ConcretePolygon copyPolygon(Polygon p1) {
		ConcretePolygon newPoly = new ConcretePolygon();
		ConcretePolygon original = p1.getOriginal();
		LinearRing ext = new LinearRing(LinearRingType.EXTERIOR);
		newPoly.setExteriorRing(ext);
		for (Vertex v : original.getExteriorRing().getVertices()) {
			ext.addVertex(v);
		}
		for (LinearRing lr : original.getInnerRings()) {
			LinearRing inner = new LinearRing(LinearRingType.INTERIOR);
			newPoly.addInteriorRing(inner);
			for (Vertex v : lr.getVertices()) {
				inner.addVertex(v);
			}
		}
		return newPoly;
	}

	private ConcretePolygon createPolygon(List<LinearRing> newRings, LinearRing lr, ConcretePolygon p) {
		lr.setType(LinearRingType.EXTERIOR);
		p.setExteriorRing(lr);
		for (LinearRing innerRing : newRings) {
			if (lr == innerRing) {
				continue;
			}
			p.addInteriorRing(innerRing);
		}
		return p;
	}

	private boolean isRingExterior(List<LinearRing> newRings, LinearRing lr) {
		for (LinearRing other : newRings) {
			if (lr == other) {
				continue;
			}
			for (Vertex v : lr.getVertices()) {
				if (other.isPointInside(v)) {
					return false;
				}
			}
		}
		return true;
	}

	private LinearRing createRingFromEdges(List<Edge> edges) {
		Edge startEdge = edges.get(0);
		Vertex start = startEdge.getFrom();
		Vertex currentVertex = startEdge.getTo();
		edges.remove(0);
		LinearRing lr = new LinearRing(LinearRingType.INTERIOR);
		lr.addVertex(start);
		lr.addVertex(currentVertex);
		while (!edges.isEmpty()) {
			boolean removedEdge = false;
			for (Edge e : edges) {
				Vertex oppositeV = e.getOppositeVertex(currentVertex);
				if (oppositeV != null) {
					lr.addVertex(oppositeV);
					currentVertex = oppositeV;
					edges.remove(e);
					removedEdge = true;
					break;
				}
			}
			if (!removedEdge) {
				// avoid endless loop in case of errors
				logger.error("Could not create ring from edges, edges are not a loop");
				break;
			}
		}
		return lr;
	}

	private void filterDuplicateEdges(List<Edge> edges, Set<Edge> duplicateEdges, List<Edge> newEdges) {
		for (Edge e : edges) {
			if (!duplicateEdges.contains(e)) {
				newEdges.add(e);
			}
		}
	}

	private List<List<Edge>> findCyclesInEdges(List<Edge> edges) {
		EdgeGraph graph = new EdgeGraph();
		for (Edge e : edges) {
			graph.addEdge(e, edges);
		}
		return graph.getCycles();
	}

	private boolean isPolygonInPlane(Plane plane, Polygon potentialPoly) {
		for (Vertex v : potentialPoly.getExteriorRing().getVertices()) {
			if (plane.getDistance(v) > EPSILON) {
				return false;
			}
		}
		return true;
	}

	@Override
	public GeometrySimplifier createNew() {
		return new GeometrySimplifier();
	}

}
